/*-
 * Copyright (c) 2012 Spectra Logic Corporation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/cdefs.h>
__FBSDID("$FreeBSD$");

#include <sys/param.h>

#ifdef _KERNEL
#include <sys/systm.h>
#include <sys/kernel.h>
#else
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#endif

#include <sys/acl.h>
#include <sys/endian.h>
#include <sys/malloc.h>
#include <sys/security_id.h>
#include <sys/security_descriptor.h>

#ifdef _KERNEL
MALLOC_DEFINE(M_SDESC, "sdesc", "Security Descriptors");
static inline void *
sd_alloc(size_t size, int flags)
{
	return (malloc(size, M_SDESC, flags));
}

static inline void *
sd_realloc(void *addr, size_t size, int flags)
{
	return (realloc(addr, size, M_SDESC, flags));
}

static inline void
sd_free(void *addr)
{
	free(addr, M_SDESC);
}
#else
static inline void *
sd_alloc(size_t size, int flags)
{
	void *space;

	space = malloc(size);
	if (space != NULL && (flags & M_ZERO) != 0)
		memset(space, 0, size);
	return (space);
}

static inline void *
sd_realloc(void *addr, size_t newsize, int flags __unused)
{
	return (realloc(addr, newsize));
}

static inline void
sd_free(void *addr)
{
	free(addr);
}
#endif

/*====================== SID Based ACL Memory Management =====================*/
struct sid_acl_entry *
sid_ace_alloc(struct sid *sid)
{
	struct sid_acl_entry *new_ace;
	size_t ace_size;

	ace_size = sizeof(*new_ace);
	if (sid != NULL)
		ace_size += sid_len(sid);

	new_ace = sd_alloc(ace_size, M_ZERO|M_WAITOK);
	new_ace->sae_size = ace_size;
	if (sid != NULL)
		memcpy(new_ace->sae_sid, sid, sid_len(sid));

	return (new_ace);
}

void
sid_ace_free(struct sid_acl_entry *ace)
{
	sd_free(ace);
}

struct sid_acl *
sid_acl_alloc(void)
{
	struct sid_acl *acl;

	acl = sd_alloc(sizeof(struct sid_acl), M_ZERO|M_WAITOK);
	if (acl != NULL)
		acl->sacl_size = sizeof(*acl);
	acl->sacl_rev = SID_ACL_REV;
	return (acl);
}

void
sid_acl_free(struct sid_acl *acl)
{
	sd_free(acl);
}

/*========================== SID Based ACL Iterator ==========================*/
void
sid_acl_iter_init(struct sid_acl_iter *iter, struct sid_acl *sacl)
{

	iter->sai_acl = sacl;
	iter->sai_next_ace_idx = 0;
	if (sacl->sacl_ace_cnt == 0)
		iter->sai_next_ace = NULL;
	else
		iter->sai_next_ace = (struct sid_acl_entry *)&sacl->sacl_data;
}

struct sid_acl_entry *
sid_acl_iter_next(struct sid_acl_iter *iter)
{
	struct sid_acl_entry *next_ace;	
	uint8_t *next_ace_data;

	if (iter->sai_next_ace_idx >= iter->sai_acl->sacl_ace_cnt)
		return (NULL);

	next_ace = iter->sai_next_ace;
	next_ace_data = (uint8_t *)next_ace;
	next_ace_data += next_ace->sae_size;
	iter->sai_next_ace = (struct sid_acl_entry *)next_ace_data;
	iter->sai_next_ace_idx++;

	return (next_ace);
}

struct sid_acl_entry *
sid_acl_iter_seek(struct sid_acl_iter *iter, uint32_t idx)
{
	struct sid_acl_entry *next_ace;	

	if (idx >= iter->sai_acl->sacl_ace_cnt) {
		iter->sai_next_ace_idx = iter->sai_acl->sacl_ace_cnt;
		iter->sai_next_ace = NULL;
		return (NULL);
	}

	if (idx < iter->sai_next_ace_idx)
		sid_acl_iter_init(iter, iter->sai_acl);

	do {
		next_ace = sid_acl_iter_next(iter);
	} while (iter->sai_next_ace_idx <= idx);

	return (next_ace);
}

/*================== SID Based ACL Construction/Modification =================*/
int
sid_acl_insert_ace(struct sid_acl **pacl, struct sid_acl_entry *ace,
    uint32_t idx)
{
	struct sid_acl *acl;
	uint8_t *ace_insert_point;
	size_t newsize;

	acl = *pacl;
	if (idx > acl->sacl_ace_cnt)
		return (EINVAL);

	newsize = acl->sacl_size + ace->sae_size;
	acl = sd_realloc(acl, newsize, M_WAITOK);

	if (idx == acl->sacl_ace_cnt) {
		ace_insert_point = (uint8_t *)acl + acl->sacl_size;
	} else {
		struct sid_acl_iter iter;

		sid_acl_iter_init(&iter, acl);
		ace_insert_point = (uint8_t *)sid_acl_iter_seek(&iter, idx);
		memmove(ace_insert_point + ace->sae_size, ace_insert_point,
		    acl->sacl_size - (ace_insert_point - (uint8_t *)acl));
	}
	memcpy(ace_insert_point, ace, ace->sae_size);
	acl->sacl_size = newsize;
	acl->sacl_ace_cnt++;
	*pacl = acl;
	return (0);
}

int
sid_acl_remove_ace(struct sid_acl **pacl, uint32_t idx)
{
	struct sid_acl_iter iter;
	struct sid_acl *acl;
	struct sid_acl_entry *ace_to_remove;
	uint8_t *ace_removal_point;
	uint8_t *next_ace;
	uint32_t change_in_size;

	acl = *pacl;
	if (idx >= acl->sacl_ace_cnt)
		return (EINVAL);

	sid_acl_iter_init(&iter, acl);
	ace_to_remove = sid_acl_iter_seek(&iter, idx);
	change_in_size = ace_to_remove->sae_size;
	ace_removal_point = (uint8_t *)ace_to_remove;
	next_ace = ace_removal_point + change_in_size;
	memmove(ace_removal_point, next_ace,
	    acl->sacl_size - (next_ace - (uint8_t *)acl));
	acl->sacl_size -= change_in_size;
	acl->sacl_ace_cnt--;

	/* Reclaim unused space. */
	*pacl = sd_realloc(acl, acl->sacl_size, M_WAITOK);
	return (0);
}

int
sid_acl_replace_ace(struct sid_acl **pacl, struct sid_acl_entry *ace,
    uint32_t idx)
{
	struct sid_acl_iter iter;
	struct sid_acl *acl;
	struct sid_acl_entry *ace_to_replace;
	uint8_t *ace_replacement_point;
	uint8_t *next_ace;
	int32_t change_in_size;

	acl = *pacl;
	if (idx >= acl->sacl_ace_cnt)
		return (EINVAL);

	sid_acl_iter_init(&iter, acl);
	ace_to_replace = sid_acl_iter_seek(&iter, idx);
	change_in_size = ace->sae_size - ace_to_replace->sae_size;

	if (change_in_size > 0) {
		acl = sd_realloc(acl, acl->sacl_size + change_in_size,
		    M_WAITOK);

		*pacl = acl;
		sid_acl_iter_init(&iter, acl);
		ace_to_replace = sid_acl_iter_seek(&iter, idx);
	}

	ace_replacement_point = (uint8_t *)ace_to_replace;
	if (change_in_size != 0) {
		next_ace = ace_replacement_point + ace_to_replace->sae_size;
		memmove(next_ace + change_in_size, next_ace,
		    acl->sacl_size - (next_ace - (uint8_t *)acl));
		acl->sacl_size += change_in_size;
	}

	memcpy(ace_replacement_point, ace, ace->sae_size);

	/* Reclaim unused space. */
	if (change_in_size < 0)
		*pacl = sd_realloc(acl, acl->sacl_size, M_WAITOK);
	return (0);
}

/*=================== Security Descriptor Memory Management ==================*/
void
security_descriptor_init(struct security_descriptor *sd)
{
	memset(sd, 0, sizeof(*sd));
	sd->sd_rev = SD_REV;
}

struct security_descriptor *
security_descriptor_alloc(void)
{
	struct security_descriptor *sd;

	sd = sd_alloc(sizeof(struct security_descriptor), M_WAITOK);
	if (sd != NULL)
		security_descriptor_init(sd);

	return (sd);
}

void
security_descriptor_free(struct security_descriptor *sd)
{
	if (sd == NULL)
		return;

	if (sd->sd_owner != NULL) {
		sid_free(sd->sd_owner);
		sd->sd_owner = NULL;
	}

	if (sd->sd_group != NULL) {
		sid_free(sd->sd_group);
		sd->sd_group = NULL;
	}

	if (sd->sd_dacl != NULL) {
		sid_acl_free(sd->sd_dacl);
		sd->sd_dacl = NULL;
	}

	if (sd->sd_sacl != NULL) {
		sid_acl_free(sd->sd_sacl);
		sd->sd_sacl = NULL;
	}

	sd_free(sd);
}

void
serialized_security_descriptor_free(struct security_descriptor_serialized *ssd)
{
	if (ssd != NULL)
		sd_free(ssd);
}

/*===================== Security Descriptor Serialization ====================*/
static int
security_descriptor_serialize_sid(struct security_descriptor_serialized *sds,
    uint32_t *sid_offset_field, struct sid *sid)
{
	if (sid == NULL) {
		*sid_offset_field = 0;
		return (0);
	}
	memcpy((uint8_t *)sds + sds->sds_size, sid, sid_len(sid));
	*sid_offset_field = sds->sds_size;
	sds->sds_size += sid_len(sid);
	return (0);
}

static int
security_descriptor_serialize_acl(struct security_descriptor_serialized *sds,
    uint32_t *acl_offset_field, struct sid_acl *acl)
{
	if (acl == NULL) {
		*acl_offset_field = 0;
		return (0);
	}
	memcpy((uint8_t *)sds + sds->sds_size, acl, acl->sacl_size);
	*acl_offset_field = sds->sds_size;
	sds->sds_size += acl->sacl_size;
	return (0);
}

int
security_descriptor_serialize(struct security_descriptor *sd,
    struct security_descriptor_serialized **psds)
{
	struct	security_descriptor_serialized *sds;
	size_t	sds_buf_size;
	int	error;

	sds_buf_size = sizeof(*sds);
	if (sd->sd_owner != NULL)
		sds_buf_size += sid_len(sd->sd_owner);
	if (sd->sd_group != NULL)
		sds_buf_size += sid_len(sd->sd_group);
	if (sd->sd_dacl != NULL)
		sds_buf_size += sd->sd_dacl->sacl_size;
	if (sd->sd_sacl != NULL)
		sds_buf_size += sd->sd_sacl->sacl_size;

	sds = sd_alloc(sds_buf_size, M_WAITOK);
	sds->sds_rev = sd->sd_rev;
	sds->sds_pad = sd->sd_pad;
	sds->sds_control = sd->sd_control;
	sds->sds_size = sizeof(*sds);

	error = security_descriptor_serialize_sid(sds, &sds->sds_owner_offset,
	    sd->sd_owner);
	if (error != 0)
		goto out;

	error = security_descriptor_serialize_sid(sds, &sds->sds_group_offset,
	    sd->sd_group);
	if (error != 0)
		goto out;

	error = security_descriptor_serialize_acl(sds, &sds->sds_dacl_offset,
	    sd->sd_dacl);
	if (error != 0)
		goto out;

	error = security_descriptor_serialize_acl(sds, &sds->sds_sacl_offset,
	    sd->sd_sacl);
	if (error != 0)
		goto out;

	sds->sds_control |= SD_CTRL_SELF_RELATIVE;
	*psds = sds;
	return (0);
out:
	sd_free(sds);
	return (error);
}

static inline struct sid *
sid_from_offset(struct security_descriptor_serialized *sds, uint32_t offset)
{
	if (offset == 0)
		return (NULL);
	return ((struct sid *)((uint8_t *)sds + offset));
}

static inline struct sid_acl *
acl_from_offset(struct security_descriptor_serialized *sds, uint32_t offset)
{
	if (offset == 0)
		return (NULL);
	return ((struct sid_acl *)((uint8_t *)sds + offset));
}

void
security_descriptor_deserialize(struct security_descriptor *sd,
    struct security_descriptor_serialized *sds)
{
	sd->sd_rev = sds->sds_rev;
	sd->sd_pad = sds->sds_pad;
	sd->sd_control = sds->sds_control & ~SD_CTRL_SELF_RELATIVE;
	sd->sd_owner = sid_from_offset(sds, sds->sds_owner_offset);
	sd->sd_group = sid_from_offset(sds, sds->sds_group_offset);
	sd->sd_dacl = acl_from_offset(sds, sds->sds_dacl_offset);
	sd->sd_sacl = acl_from_offset(sds, sds->sds_sacl_offset);
}
