/*-
 * Copyright (c) 2012 Spectra Logic Corporation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/cdefs.h>
__FBSDID("$FreeBSD$");

#include <sys/param.h>

#ifdef _KERNEL
#include <sys/systm.h>
#include <sys/kernel.h>
#else
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#endif

#include <sys/endian.h>
#include <sys/malloc.h>
#include <sys/security_id.h>

#include <machine/_inttypes.h>

#ifdef _KERNEL
MALLOC_DEFINE(M_SID, "sid", "Security Identifiers");
static inline void *
sid_alloc(size_t size, int flags)
{
	return (malloc(size, M_SID, flags));
}
#else
static inline void *
sid_alloc(size_t size, int flags)
{
	void *space;

	space = malloc(size);
	if (space != NULL && (flags & M_ZERO) != 0)
		memset(space, 0, size);
	return (space);
}
#endif

static int
sid_from_string_pad(struct sid **psid, const char *sid_string,
    int pad_authorities)
{
	uint64_t sid_id_authority;
	uint64_t sid_sub_authority;
	struct	 sid *sid;
	size_t	 sid_sz;
	int	 sid_rev;
	int	 sub_authorities_start;
	int	 sub_authority_count;
	int	 conversions;
	int	 i;
	int	 j;

	*psid = NULL;

	/* Don't count version or primary authority. */
	sub_authority_count = -2;
	for (i = 0; sid_string[i] != '\0'; i++) {
		if (sid_string[i] == '-') {
			sub_authority_count++;
		
			/* All values are unsigned. */
			if (i != 0 && sid_string[i - 1] == '-')
				return (EINVAL);
		}
	}

	conversions = sscanf(sid_string, "S-%3d-%14"SCNi64"%n",
	    &sid_rev, &sid_id_authority, &sub_authorities_start);
	if (conversions != 2)
		return (EINVAL);

	if (sid_rev > 255 || sid_id_authority > 0xFFFFFFFFFFFF)
		return (EOVERFLOW);

	sid_sz = sizeof(struct sid)
	    + ((sub_authority_count + pad_authorities) * sizeof(uint32_t));
	sid = sid_alloc(sid_sz, M_WAITOK|M_ZERO);
	sid->sid_rev = sid_rev;
	sid->sid_sub_authority_count = sub_authority_count;
	sid_set_id_authority(sid, sid_id_authority);

	sid_string += sub_authorities_start;
	for (j = 0; j < sub_authority_count; j++, sid_string += i) {

		/* No leading 0's or "0x". */
		if (sid_string[1] == '0'
		 && sid_string[2] != '\0' && sid_string[2] != '-') {
			sid_free(sid);
			return (EINVAL);
		}

		conversions = sscanf(sid_string, "-%10"SCNu64"%n",
			&sid_sub_authority, &i);
		if (conversions != 1) {
			sid_free(sid);
			return (EINVAL);
		}

		if (sid_string[i] != '\0' && sid_string[i] != '-')
			return (EINVAL);

		if (sid_sub_authority > 0xFFFFFFFF) {
			sid_free(sid);
			return (EOVERFLOW);
		}

		sid_set_sub_authority(sid, j, sid_sub_authority & 0xFFFFFFFF);
	}	
	*psid = sid;
	return (0);
}

int
sid_from_string(struct sid **psid, const char *sid_string)
{
	return (sid_from_string_pad(psid, sid_string, /*pad*/0));
}

int
sid_from_domain_and_rid(struct sid **psid, const char *sid_string,
    uint32_t rid)
{
	int error;

	error = sid_from_string_pad(psid, sid_string, /*pad*/1);
	if (error != 0)
		return (error);

	sid_set_sub_authority(*psid, (*psid)->sid_sub_authority_count, rid);
	(*psid)->sid_sub_authority_count++;
	return (0);
}

static int
sid_to_string_limited(const struct sid *sid, char **pstring_sid,
    int sub_authority_limit)
{
	uint64_t sid_id_authority;
	char	*string_sid;
	size_t   string_sz;
	int	 i;

	if (sid->sid_sub_authority_count == 0)
		return (EINVAL);

	if (sub_authority_limit > sid->sid_sub_authority_count)
		sub_authority_limit = sid->sid_sub_authority_count;

	string_sz = sizeof("S-255-0xFFFFFFFFFFFF"); /* includes final NUL */
	string_sz += sid->sid_sub_authority_count * (sizeof("-4294967295") - 1);

	string_sid = sid_alloc(string_sz, M_WAITOK);
	*pstring_sid = string_sid;
	sid_id_authority = sid_get_id_authority(sid);
	if (sid_id_authority < 0xFFFFFFFF)
		string_sid += sprintf(string_sid, "S-%d-%"PRIu64,
		    sid->sid_rev, sid_id_authority);
	else
		string_sid += sprintf(string_sid, "S-%d-0x%.12"PRIX64,
		    sid->sid_rev, sid_id_authority);

	for (i = 0; i < sub_authority_limit; i++)
		string_sid += sprintf(string_sid, "-%"PRIu32,
		    sid_get_sub_authority(sid, i));

	return (0);
}

int
sid_to_string(const struct sid *sid, char **pstring_sid)
{
	return (sid_to_string_limited(sid, pstring_sid,
	    sid->sid_sub_authority_count));
}

int
sid_to_domain_and_rid(const struct sid *sid, char **pdomain, uint32_t *prid)
{
	int error;

	error = sid_to_string_limited(sid, pdomain,
	    sid->sid_sub_authority_count - 1);
	if (error != 0)
		return (error);
	*prid = sid_get_sub_authority(sid, sid->sid_sub_authority_count - 1);
	return (0);
}
