/*-
 * Copyright (c) 2012 Spectra Logic Corporation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifdef _KERNEL
#include <sys/cdefs.h>
__FBSDID("$FreeBSD$");
#endif /* !_KERNEL */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/sysproto.h>
#include <sys/capability.h>
#include <sys/fcntl.h>
#include <sys/filedesc.h>
#include <sys/kernel.h>
#include <sys/malloc.h>
#include <sys/vnode.h>
#include <sys/lock.h>
#include <sys/mount.h>
#include <sys/mutex.h>
#include <sys/namei.h>
#include <sys/proc.h>
#include <sys/acl.h>
#include <sys/security_id.h>
#include <sys/security_descriptor.h>

#include <security/mac/mac_framework.h>

static int
security_descriptor_get(struct thread *td, struct vnode *vp, int secinfo,
    struct security_descriptor **psd)
{
	int error;

	vn_lock(vp, LK_EXCLUSIVE | LK_RETRY);
#ifdef MAC
	error = mac_vnode_check_getsd(td->td_ucred, vp, secinfo);
	if (error != 0)
		goto out;
#endif
	error = VOP_GETSD(vp, secinfo, psd, td->td_ucred, td);

#ifdef MAC
out:
#endif
	VOP_UNLOCK(vp, 0);
	return (error);
}

static int
security_descriptor_set(struct thread *td, struct vnode *vp, int secinfo,
    struct security_descriptor *sd)
{
	int error;

	vn_lock(vp, LK_EXCLUSIVE | LK_RETRY);
#ifdef MAC
	error = mac_vnode_check_setsd(td->td_ucred, vp, secinfo, sd);
	if (error != 0)
		goto out;
#endif
	error = VOP_SETSD(vp, secinfo, sd, td->td_ucred, td);

#ifdef MAC
out:
#endif
	VOP_UNLOCK(vp, 0);
	return (error);
}

int    
sys_security_descriptor_get_fd(struct thread *td,
    struct security_descriptor_get_fd_args *uap)
{
	struct file *fp;
	struct security_descriptor *sd;
	struct security_descriptor_serialized *sds;
	cap_rights_t rights;
	int error;

	error = getvnode(td->td_proc->p_fd, uap->fd,
	    cap_rights_init(&rights, CAP_SD_GET), &fp);
	if (error != 0)
		return (error);

	sd = NULL;
	sds = NULL;
	error = security_descriptor_get(td, fp->f_vnode,
	    uap->security_info, &sd);
	fdrop(fp, td);

	if (error == 0)
		error = security_descriptor_serialize(sd, &sds);

	if (error == 0 && uap->sd != NULL)
		error = copyout(sds, uap->sd,
		    MIN(uap->sd_buf_size, sds->sds_size));

	if (error == 0 && uap->sd_size != NULL) {
		size_t sds_size;

		sds_size = sds->sds_size;
		error = copyout(&sds_size, uap->sd_size, sizeof(*uap->sd_size));
	}

	security_descriptor_free(sd);
	serialized_security_descriptor_free(sds);
	return (error);
}

int
sys_security_descriptor_set_fd(struct thread *td,
    struct security_descriptor_set_fd_args *uap)
{
	struct security_descriptor sd;
	struct security_descriptor_serialized *sds;
	cap_rights_t rights;
	struct file *fp;
	int error;

	/* Copyin */
	sds = malloc(uap->sd_size, M_SDESC, M_WAITOK);
	error = copyin(uap->sd, sds, uap->sd_size);
	if (error != 0)
		goto out;

	error = getvnode(td->td_proc->p_fd, uap->fd,
	    cap_rights_init(&rights, CAP_SD_SET), &fp);
	if (error != 0)
		goto out;

	security_descriptor_deserialize(&sd, sds);

	error = security_descriptor_set(td, fp->f_vnode,
	    uap->security_info, &sd);
	fdrop(fp, td);

out:
	serialized_security_descriptor_free(sds);
	return (error);
}

int    
sys_security_descriptor_get_file(struct thread *td,
    struct security_descriptor_get_file_args *uap)
{
	struct nameidata nd;
	int error;
	struct security_descriptor *sd;
	struct security_descriptor_serialized *sds;

	NDINIT(&nd, LOOKUP, FOLLOW, UIO_USERSPACE, uap->path, td);
	error = namei(&nd);
	if (error != 0)
		return (error);

	sd = NULL;
	sds = NULL;
	error = security_descriptor_get(td, nd.ni_vp, uap->security_info, &sd);
	NDFREE(&nd, 0);

	if (error == 0)
		error = security_descriptor_serialize(sd, &sds);

	if (error == 0 && uap->sd != NULL)
		error = copyout(sds, uap->sd,
		    MIN(uap->sd_buf_size, sds->sds_size));

	if (error == 0 && uap->sd_size != NULL) {
		size_t sds_size;

		sds_size = sds->sds_size;
		error = copyout(&sds_size, uap->sd_size, sizeof(*uap->sd_size));
	}

	security_descriptor_free(sd);
	serialized_security_descriptor_free(sds);
	return (error);
}

int
sys_security_descriptor_set_file(struct thread *td,
    struct security_descriptor_set_file_args *uap)
{
	struct nameidata nd;
	struct security_descriptor sd;
	struct security_descriptor_serialized *sds;
	int error;

	NDINIT(&nd, LOOKUP, FOLLOW, UIO_USERSPACE, uap->path, td);
	error = namei(&nd);
	if (error != 0)
		return (error);

	/* Copyin */
	sds = malloc(uap->sd_size, M_SDESC, M_WAITOK);
	error = copyin(uap->sd, sds, uap->sd_size);
	if (error != 0) {
		serialized_security_descriptor_free(sds);
		NDFREE(&nd, 0);
		return (error);
	}

	security_descriptor_deserialize(&sd, sds);

	error = security_descriptor_set(td, nd.ni_vp, uap->security_info, &sd);
	NDFREE(&nd, 0);
	serialized_security_descriptor_free(sds);
	return (error);
}
