/*-
 * Copyright (c) 2012 Spectra Logic Corporation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions, and the following disclaimer,
 *    without modification, immediately at the beginning of the file.
 * 2. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $FreeBSD$
 */

#ifndef _SYS_SECURITY_DESCRIPTOR_H_
#define	_SYS_SECURITY_DESCRIPTOR_H_

#ifdef _KERNEL
MALLOC_DECLARE(M_SDESC);
#else
#include <sys/cdefs.h>
#endif

/**
 * \brief Security ID based Access Control Entry - Identical to FreeBSD's
 *        standard ACE type with the exception that the principal is recorded
 *        by SID instead of UID/GID.
 *
 * \note  This structure must be sized and aligned to a 32bit boundary in
 *        order for it to be properly embedded into a sid_acl.
 */
struct sid_acl_entry {
	acl_tag_t		sae_tag;
	acl_perm_t		sae_perm;
	/* NFSv4 entry type, "allow" or "deny".  Unused in POSIX.1e ACLs. */
	acl_entry_type_t	sae_entry_type;
	/* NFSv4 ACL inheritance.  Unused in POSIX.1e ACLs. */
	acl_flag_t		sae_flags;
	uint32_t		sae_size;
	/* At most one, but may be empty. */
	struct sid		sae_sid[];
};

enum {
	/**
	 * Revision of ACLs that do not contain any object based ACEs.
	 *
	 * \todo  This convention is taken from Windows and should
	 *        be revisited once object based ACE support is added
	 *        to FreeBSD's implementation.
	 */
	SID_ACL_REV = 2,
	/**
	 * Revision of ACLs that contain object based ACEs.
	 *
	 * \todo  This convention is taken from Windows and should
	 *        be revisited once object based ACE support is added
	 *        to FreeBSD's implementation.
	 */
	SID_DS_ACL_REV = 4
};

/**
 * \brief Security ID based Access Control List - A variable length array of
 *        sid_acl_entry objects.
 *
 * \note  As the size of sid_acl_entry objects varies depending on the SID
 *        encoded, the sid_acl_iter APIs should be used to enumerate the
 *        contents of a sid_acl.
 *
 * \note  This structure must be sized and aligned to a 32bit boundary in
 *        order for it to be properly embedded into a
 *        security_descriptor_serialized object.
 */
struct sid_acl {
	uint32_t sacl_rev;
	uint32_t sacl_size;
	uint32_t sacl_ace_cnt;
	uint32_t sacl_data[];
};

/**
 * \brief Object used to track iteration state on a Security ID based
 *        access controll list.
 */
struct sid_acl_iter {
	struct sid_acl *sai_acl;
	struct sid_acl_entry *sai_next_ace;
	uint32_t sai_next_ace_idx;
};

enum {
	/** In Security Descriptor APIs, request/set owner info. */
	SD_INFO_OWNER = 0x00000001,
	/** In Security Descriptor APIs, request/set group info. */
	SD_INFO_GROUP = 0x00000002,
	/**
	 * In Security Descriptor APIs, request/set discretionary
	 * access control list data.
	 */
	SD_INFO_DACL  = 0x00000004,
	/**
	 * In Security Descriptor APIs, request/set system
	 * access control list data.
	 */
	SD_INFO_SACL  = 0x00000008
};

enum {
	SD_CTRL_OWNER_DEFAULTED		= 0x0001,
	SD_CTRL_GROUP_DEFAULTED		= 0x0002,
	SD_CTRL_DACL_PRESENT		= 0x0004,
	SD_CTRL_DACL_DEFAULTED		= 0x0008,
	SD_CTRL_SACL_PRESENT		= 0x0010,
	SD_CTRL_SACL_DEFAULTED		= 0x0020,
	SD_CTRL_DACL_UNTRUSTED		= 0x0040,
	SD_CTRL_SERVER_SECURITY		= 0x0080,
	SD_CTRL_DACL_AUTO_INHERIT_REQ	= 0x0100,
	SD_CTRL_SACL_AUTO_INHERIT_REQ	= 0x0200,
	SD_CTRL_DACL_AUTO_INHERITED	= 0x0400,
	SD_CTRL_SACL_AUTO_INHERITED	= 0x0800,
	SD_CTRL_DACL_PROTECTED		= 0x1000,
	SD_CTRL_SACL_PROTECTED		= 0x2000,
	/** Resource manager information is valid in sd_pad */
	SD_CTRL_RM_CONTROL_VALID	= 0x4000,
	/**
	 * If set, security descriptor has the format of
	 * security_descriptor_serialized.
	 */
	SD_CTRL_SELF_RELATIVE		= 0x8000
};

enum {
	/**
	 * The current security descriptor and serialized security descriptor
	 * format revision.
	 */
	SD_REV = 1
};

/**
 * \brief Structure for Getting/Setting owner, group, and ACL based
 *        attributes for an object.
 *
 * When setting data within a security descriptor, the
 * security_descriptor_set_*() APIs must be used so that the
 * sd_control field is properly updated.
 */
struct security_descriptor {
	uint8_t		sd_rev;
	uint8_t		sd_pad;
	uint16_t	sd_control;
	struct sid     *sd_owner;
	struct sid     *sd_group;
	struct sid_acl *sd_dacl;
	struct sid_acl *sd_sacl;
};

/**
 * \brief Structure containing identical content to security_descriptor,
 *        but packed into a contiguous memory location.
 */
struct security_descriptor_serialized {
	uint8_t		sds_rev;
	uint8_t		sds_pad;
	uint16_t	sds_control;
	uint32_t	sds_size;
	uint32_t	sds_owner_offset;
	uint32_t	sds_group_offset;
	uint32_t	sds_dacl_offset;
	uint32_t	sds_sacl_offset;
	uint8_t		sds_data[];
};

__BEGIN_DECLS
/**
 * \brief Allocate and initialize a SID based ACL entry referencing the
 *        given SID.
 *
 * After allocation the sae_tag, sae_perm, sae_entry_type, and sae_flags
 * must be modified to reflect the desired final ACE.  For ACL_USER_OBJ,
 * ACL_GROUP_OBJ, and ACL_EVERYONE ACE entries, the specified SID must
 * be NULL.
 *
 * \note  The ACL and Security Descriptor APIs assume that all storage
 *        associated with SIDs, ACLs, and Security Descriptors are managed
 *        using the object specific memory allocation routines.  Allocating
 *        storage using a different API may result in undefined behavior.
 */
struct sid_acl_entry *sid_ace_alloc(struct sid *);

/**
 * \brief Free the contents of a SID based ACL entry.
 *
 * \note  The ACL and Security Descriptor APIs assume that all storage
 *        associated with SIDs, ACLs, and Security Descriptors are managed
 *        using the object specific memory allocation routines.  Allocating
 *        storage using a different API may result in undefined behavior.
 */
void sid_ace_free(struct sid_acl_entry *);

/**
 * \brief Initialize a SID ACL iterator so that it points to the first
 *        entry (if any) in the given ACL.
 *
 * \note  Iterators are typically declared on the stack and then
 *        initialized with this routine prior to first use.
 *
 * \note  Iterators must be reinitialized after an ACL is modified.
 */
void sid_acl_iter_init(struct sid_acl_iter *, struct sid_acl *);

/**
 * \brief Return the ACE at the iterator's current position and advance
 *        the iterator to the next position.
 *
 * \return  NULL if the ACL is empty or the iterator has passed the end
 *          of the ACL.  Otherwise the ACE at the iterator's current
 *          position.
 */
struct sid_acl_entry *sid_acl_iter_next(struct sid_acl_iter *);

/**
 * \brief Move the position of a SID ACL iterator so that it points to
 *        ACE entry at the given index and return this ACE (if any).
 */
struct sid_acl_entry *sid_acl_iter_seek(struct sid_acl_iter *, uint32_t);

/**
 * \brief Allocate and initialize a SID based Access Control List.
 *
 * Once allocated, ACE entries may be added to the ACL via the
 * sid_acl_*_ace() methods.
 *
 * \note  The ACL and Security Descriptor APIs assume that all storage
 *        associated with SIDs, ACLs, and Security Descriptors are managed
 *        using the object specific memory allocation routines.  Allocating
 *        storage using a different API may result in undefined behavior.
 */
struct sid_acl *sid_acl_alloc(void);

/*
 * \brief Free the contents of a SID based Access Control List.
 *
 * \note  The ACL and Security Descriptor APIs assume that all storage
 *        associated with SIDs, ACLs, and Security Descriptors are managed
 *        using the object specific memory allocation routines.  Allocating
 *        storage using a different API may result in undefined behavior.
 */
void sid_acl_free(struct sid_acl *);

/**
 * \brief Insert a SID based ACE at a specific location in the given
 *        SID based Access Control List.
 *
 * \note  Upon successful modification of the Access Control List, its
 *        storage location may change and any cached pointers to it become
 *        invalid.
 *
 * \note  After insertion, the ACL is self contained and the storage
 *        for the passed in ACE may be reclaimed.
 */
int sid_acl_insert_ace(struct sid_acl **, struct sid_acl_entry *, uint32_t);

/**
 * \brief Remove a SID based ACE at a specific location in the given
 *        SID based Access Control List.
 *
 * \note  Upon successful modification of the Access Control List, its
 *        storage location may change and any cached pointers to it become
 *        invalid.
 */
int sid_acl_remove_ace(struct sid_acl **, uint32_t);

/**
 * \brief Replace the contents of a SID based ACE at a specific location
 *        in the given SID based Access Control List.
 *
 * \note  Upon successful modification of the Access Control List, its
 *        storage location may change and any cached pointers to it become
 *        invalid.
 *
 * \note  After replacement, the ACL is self contained and the storage
 *        for the passed in ACE may be reclaimed.
 */
int sid_acl_replace_ace(struct sid_acl **, struct sid_acl_entry *, uint32_t);

/**
 * \brief Initialize a security descriptor to the empty state.
 *
 * This API is intended for use in initializing a security descriptor
 * allocated on the stack.  This type of security descriptor is typically
 * used as a staging area for owner/group SIDs and ACLs that have been
 * allocated using external APIs (e.g. Samba/Likewise) so they can be
 * passed to security_descriptor_serialize(), or for the reverse process
 * of interrogating the content of a serialized security descriptor.
 * security_descriptor_free() must never be called in these situations
 * as the storage referenced by the security descriptor is not owned
 * by it.
 */
void security_descriptor_init(struct security_descriptor *);

/**
 * \brief Allocate and initialize a security descriptor to the empty state.
 *
 * \note  The ACL and Security Descriptor APIs assume that all storage
 *        associated with SIDs, ACLs, and Security Descriptors are managed
 *        using the object specific memory allocation routines.  Allocating
 *        storage using a different API may result in undefined behavior.
 */
struct security_descriptor *security_descriptor_alloc(void);

/**
 * \brief Free a security descriptor and all SIDs and ACLs it references.
 *
 * \note  The ACL and Security Descriptor APIs assume that all storage
 *        associated with SIDs, ACLs, and Security Descriptors are managed
 *        using the object specific memory allocation routines.  Allocating
 *        storage using a different API may result in undefined behavior.
 */
void security_descriptor_free(struct security_descriptor *);

/**
 * \brief Free a serialized security descriptor and all SIDs and ACLs
 *        it references.
 *
 * \note  The ACL and Security Descriptor APIs assume that all storage
 *        associated with SIDs, ACLs, and Security Descriptors are managed
 *        using the object specific memory allocation routines.  Allocating
 *        storage using a different API may result in undefined behavior.
 */
void serialized_security_descriptor_free(
    struct security_descriptor_serialized *);

/**
 * \brief Allocate space for and initialized the contents of a
 *        serialized security descriptor with the contents of a
 *        non-serialized security descriptor.
 *
 * \note  The ACL and Security Descriptor APIs assume that all storage
 *        associated with SIDs, ACLs, and Security Descriptors are managed
 *        using the object specific memory allocation routines.  Allocating
 *        storage using a different API may result in undefined behavior.
 *
 * \note  Upon successful creation the serialized security descriptor
 *        is self contained and the storage for the passed in security
 *        descriptor may be reclaimed.
 */
int security_descriptor_serialize(struct security_descriptor *,
    struct security_descriptor_serialized **);

/**
 * \brief Initialized the contents of a security descriptor with
 *        references to the data in the specified serialized security
 *        descriptor.
 *
 * \note  The storage for the serialized security descriptor can only
 *        be reclaimed after all references to deserialized data structure
 *        have completed.
 *
 * \note  The deserialized security descriptor does not own the data it
 *        references and thus must never be passed to
 *        security_descriptor_free().  The typical use case is to allocate
 *        a security_descriptor on the stack and allow its storage to
 *        be reclaimed automatically once it passes out of scope.
 */
void security_descriptor_deserialize(struct security_descriptor *,
    struct security_descriptor_serialized *);

#ifndef _KERNEL
/**
 * \brief Retrieve owner/group/ACL information on a file object referenced
 *        by file descriptor.
 *
 * \param _fd             The file to interrogate.
 * \param _security_info  A bit mask of attributes to fetch.
 * \param _sds            A buffer for reception of the requested information.
 *                        The data will be formated as a
 *                        security_descriptor_serialized object.  This
 *                        argument may be NULL if only requesting the
 *                        size of the buffer necessary to fully contain
 *                        the requested data.
 * \param _sd_buf_size    The size of the buffer provided in _sds.  Must be
 *                        zero if _sds is NULL.
 * \param _sd_size        The returned buffer size required to fully contain
 *                        the requested data.  May be NULL.
 */
int security_descriptor_get_fd(int _fd, int _security_info,
    struct security_descriptor_serialized *_sds, size_t _sd_buf_size,
    size_t *_sd_size);

/**
 * \brief Set owner/group/ACL information on a file object referenced
 *        by file descriptor.
 *
 * \param _fd             The file to modify.
 * \param _security_info  A bit mask of attributes to set.
 * \param _sds            A buffer containing the data to set, formatted
 *                        as a security_descriptor_serialized object.
 * \param _sd_size        The size of the buffer provided in _sds.
 */
int security_descriptor_set_fd(int _fd, int _security_info,
    struct security_descriptor_serialized *_sds, size_t _sd_size);

/**
 * \brief Retrieve owner/group/ACL information on a file object referenced
 *        by file name.
 *
 * \param _path           The path to the file to interrogate.
 * \param _security_info  A bit mask of attributes to fetch.
 * \param _sds            A buffer for reception of the requested information.
 *                        The data will be formated as a
 *                        security_descriptor_serialized object.  This
 *                        argument may be NULL if only requesting the
 *                        size of the buffer necessary to fully contain
 *                        the requested data.
 * \param _sd_buf_size    The size of the buffer provided in _sds.  Must be
 *                        zero if _sds is NULL.
 * \param _sd_size        The returned buffer size required to fully contain
 *                        the requested data.  May be NULL.
 */
int security_descriptor_get_file(const char *_path, int _security_info,
    struct security_descriptor_serialized *_sds, size_t _sd_buf_size,
    size_t *_sd_size);

/**
 * \brief Set owner/group/ACL information on a file object referenced
 *        by file name.
 *
 * \param _fd             The path to the file to modify.
 * \param _security_info  A bit mask of attributes to set.
 * \param _sds            A buffer containing the data to set, formatted
 *                        as a security_descriptor_serialized object.
 * \param _sd_size        The size of the buffer provided in _sds.
 */
int security_descriptor_set_file(const char *_path, int _security_info,
    struct security_descriptor_serialized *_sds, size_t _sd_size);
#endif
__END_DECLS

/**
 * \brief Append a SID based ACE to the end of the given SID based Access
 *        Control List.
 *
 * \note  Upon successful modification of the Access Control List, its
 *        storage location may change and any cached pointers to it become
 *        invalid.
 *
 * \note  After insertion, the ACL is self contained and the storage
 *        for the passed in ACE may be reclaimed.
 */
static inline int
sid_acl_append_ace(struct sid_acl **_pacl, struct sid_acl_entry *_ace)
{
	return (sid_acl_insert_ace(_pacl, _ace, (*_pacl)->sacl_ace_cnt));
}

static inline void
security_descriptor_set_owner(struct security_descriptor *_sd,
    struct sid *_sid, int _defaulted)
{
	_sd->sd_owner = _sid;
	_sd->sd_control &= ~SD_CTRL_OWNER_DEFAULTED;
	if (_sid != NULL && _defaulted != 0)
		_sd->sd_control |= ~SD_CTRL_OWNER_DEFAULTED;
}

static inline void
security_descriptor_set_group(struct security_descriptor *_sd,
    struct sid *_sid, int _defaulted)
{
	_sd->sd_group = _sid;
	_sd->sd_control &= ~SD_CTRL_GROUP_DEFAULTED;
	if (_sid != NULL && _defaulted != 0)
		_sd->sd_control |= ~SD_CTRL_GROUP_DEFAULTED;
}

static inline void
security_descriptor_set_dacl(struct security_descriptor *_sd,
    struct sid_acl *_acl, int _defaulted)
{
	_sd->sd_dacl = _acl;
	_sd->sd_control &= ~(SD_CTRL_DACL_PRESENT|SD_CTRL_DACL_DEFAULTED);
	if (_sd->sd_dacl != NULL) {
		_sd->sd_control |= SD_CTRL_DACL_PRESENT;
		if (_defaulted != 0)
			_sd->sd_control |= SD_CTRL_DACL_DEFAULTED;
	}
}

static inline void
security_descriptor_set_sacl(struct security_descriptor *_sd,
    struct sid_acl *_acl, int _defaulted)
{
	_sd->sd_sacl = _acl;
	_sd->sd_control &= ~(SD_CTRL_SACL_PRESENT|SD_CTRL_SACL_DEFAULTED);
	if (_sd->sd_sacl != NULL) {
		_sd->sd_control |= SD_CTRL_SACL_PRESENT;
		if (_defaulted != 0)
			_sd->sd_control |= SD_CTRL_SACL_DEFAULTED;
	}
}

#endif /* !_SYS_SECURITY_DESCRIPTOR_H_ */
