/*-
 * Copyright (c) 2012 Spectra Logic Corporation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions, and the following disclaimer,
 *    without modification, immediately at the beginning of the file.
 * 2. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $FreeBSD$
 */

#ifndef _SYS_SECURITY_ID_H_
#define	_SYS_SECURITY_ID_H_

#include <sys/endian.h>
#include <sys/malloc.h>

#ifdef _KERNEL
MALLOC_DECLARE(M_SID);
#endif

/**
 * \brief 48bit, Big-Endian encoded, primary authority for a security ID.
 */
struct sid_id_authority {
	uint8_t value[6];
};

#define	SID_ID_AUTHORITY_NULL		    { 0, 0, 0, 0, 0,  0 }
#define	SID_ID_AUTHORITY_WORLD		    { 0, 0, 0, 0, 0,  1 }
#define	SID_ID_AUTHORITY_LOCAL		    { 0, 0, 0, 0, 0,  2 }
#define	SID_ID_AUTHORITY_CREATOR	    { 0, 0, 0, 0, 0,  3 }
#define	SID_ID_AUTHORITY_NON_UNIQUE	    { 0, 0, 0, 0, 0,  4 }
#define	SID_ID_AUTHORITY_NT		    { 0, 0, 0, 0, 0,  5 }
#define	SID_ID_AUTHORITY_MANDATORY_LABEL    { 0, 0, 0, 0, 0, 16 }

/**
 * \brief The binary representation of a Security Identifier.
 *
 * \note  Sub-authorities are encoded in machine endian format
 *        and may need to be byte-swapped when interacting with
 *        other software components.
 */
struct sid {
	uint8_t  sid_rev;
	uint8_t  sid_sub_authority_count;
	struct	 sid_id_authority sid_id_authority;
	uint32_t sid_sub_authorities[];
};

#ifdef _KERNEL
static inline void
sid_free(struct sid *_sid)
{
	free(_sid, M_SID);
}

static inline void
sid_string_free(char *_sid_string)
{
	free(_sid_string, M_SID);
}
#else
static inline void
sid_free(struct sid *_sid)
{
	free(_sid);
}

static inline void
sid_string_free(char *_sid_string)
{
	free(_sid_string);
}
#endif /* !_KERNEL */

/**
 * \brief Calculate the allocation space requirements for a binary
 *        Security ID.
 */
static inline size_t
sid_len(const struct sid *_s)
{
	return (sizeof(*_s) +
	    sizeof(_s->sid_sub_authorities[0]) * _s->sid_sub_authority_count);
}

/**
 * \brief Convert the primary authority of a Security ID into a 64bit
 *        integer in local machine endian format.
 */
static inline uint64_t
sid_get_id_authority(const struct sid *_sid)
{
	return (((uint64_t)_sid->sid_id_authority.value[0] << 40)
	      | ((uint64_t)_sid->sid_id_authority.value[1] << 32)
	      | ((uint64_t)_sid->sid_id_authority.value[2] << 24)
	      | ((uint64_t)_sid->sid_id_authority.value[3] << 16)
	      | ((uint64_t)_sid->sid_id_authority.value[4] <<  8)
	      | ((uint64_t)_sid->sid_id_authority.value[5]));
}

/**
 * \brief Set the primary authority of a Security ID using a 64bit
 *        integer in local machine endian format.
 */
static inline void
sid_set_id_authority(struct sid *_sid, uint64_t _sid_id_authority)
{
	_sid->sid_id_authority.value[0] = (_sid_id_authority >> 40) & 0xFF;
	_sid->sid_id_authority.value[1] = (_sid_id_authority >> 32) & 0xFF;
	_sid->sid_id_authority.value[2] = (_sid_id_authority >> 24) & 0xFF;
	_sid->sid_id_authority.value[3] = (_sid_id_authority >> 16) & 0xFF;
	_sid->sid_id_authority.value[4] = (_sid_id_authority >>  8) & 0xFF;
	_sid->sid_id_authority.value[5] = (_sid_id_authority & 0xFF);
}

/**
 * \brief Retrieve the sub-authority ID at the given index as a 32bit
 *        integer in local machine endian format.
 */
static inline uint32_t
sid_get_sub_authority(const struct sid *_sid, unsigned int _idx)
{
	return (_sid->sid_sub_authorities[_idx]);
}

/**
 * \brief Set the sub-authority ID at the given index from a 32bit
 *        integer in local machine endian format.
 */
static inline void
sid_set_sub_authority(struct sid *_sid, unsigned int _idx,
    uint32_t _sid_sub_authority)
{
	_sid->sid_sub_authorities[_idx] = _sid_sub_authority;
}

/**
 * \brief Allocate and initialize a binary Security ID object from
 *        a string in Micorsoft's standard SID String format.
 *
 * \note  The storage for the returned binary Security ID must be
 *        free'd using the sid_free() api.
 */
int sid_from_string(struct sid **, const char *);

/**
 * \brief Allocate and initialize a binary Security ID object from
 *        a string in Micorsoft's standard SID String format representing
 *        all Security ID information except the final relative
 *        ID, and a 32bit relative ID in local machine endian format.
 *
 * \note  The storage for the returned binary Security ID must be
 *        free'd using the sid_free() api.
 */
int sid_from_domain_and_rid(struct sid **, const char *, uint32_t);

/**
 * \brief Allocate and initialize a string in Micorsoft's standard SID
 *	  String format that represents the given binary Security ID.
 *
 * \note  The storage for the returned string must be free'd using the
 *        sid_string_free() api.
 */
int sid_to_string(const struct sid *, char **);

/**
 * \brief Allocate and initialize a string in Micorsoft's standard SID
 *	  String format that represents all Security ID information except
 *        for the final relative ID, and extract the 32bit relative ID
 *        in local machine endian format.
 *
 * \note  The storage for the returned string must be free'd using the
 *        sid_string_free() api.
 */
int sid_to_domain_and_rid(const struct sid *, char **, uint32_t *);
#endif /* !_SYS_SECURITY_ID_H_ */
